

class SearchFiles:

    def search_files(self, directory_path: str = None, search_string: str or list = None, save_path: str = None):
        """
        This function will go through a directory and all subdirectories and try to open each file no matter
        if it is another directory, a sql code, hql, or python file, or whatever. It will search for a substring
        or list of substrings within the file. It will then create a Polars dataframe with the following columns
        (Path, Filename, Start, Occurrences, Lines, Context). If the save_path is not None it will save the
        dataframe to the save_path.

        Parameters
        ----------
        directory_path : str
            The directory to recursively search through.
        search_string : str or list
            The substring or the list of substrings is the search criteria.
        save_path : str
            The location where to save the results.

        Returns
        -------
        Polars dataframe
            A Polars dataframe with the following columns (Path, Filename, Start, Occurrences, Lines, Context).
        """

        import os
        import re
        import pandas as pd
        import json

        # Check if directory_path is valid
        if not os.path.isdir(directory_path):
            raise ValueError('directory_path is not a valid directory.')

        # Check if search_string is valid
        if not isinstance(search_string, str) and not isinstance(search_string, list):
            raise ValueError('search_string must be a string or a list of strings.')

        if isinstance(search_string, str):
            search_string = [search_string]

        # Initialize the Polars dataframe
        df = {}

        for i in search_string:
            print(i)

            df[i] = {}

            # Recursively search through the directory
            for root, dirs, files in os.walk(directory_path):
                for file in files:
                    # Get the relative path for the file
                    path = os.path.relpath(os.path.join(root, file), directory_path)
                    filepath = path
                    print(f"Searching:\t{filepath}/{file}")

                    with open(os.path.join(root, file), 'rb') as f:
                        byte_sequence = f.read()  # read the first 100 bytes
                        lines = byte_sequence.decode('iso-8859-1')

                        if f'{i.upper()} ' in lines.upper():
                            print('test')

                            indexes = [index.start() for index in re.finditer(i.upper(), lines.upper())]

                            instance = 1
                            for index in indexes:
                                if index < 100:
                                    index = 100
                                instance += 1
                                if 'pos'.upper() not in lines[index - 10: index].upper():
                                    df[i][filepath] = {}
                                    df[i][filepath]['Start'] = lines[:500]
                                    df[i][filepath]['char_index'] = {}
                                    df[i][filepath]['char_index'][str(index)] = lines[index - 100: index + 100]

            print(i)

        reformed_dict = {}
        for k, v in df.items():
            for x, y in v.items():
                reformed_dict[(k, x)] = y

        df_pd = pd.DataFrame(reformed_dict).T

        df_pd.to_csv(save_path)

        print(df_pd)

        return df_pd




from NikeSF import Snowflake
from NikeQA import QA
from Dashboards.Dashboards import Dashboards
from Dashboards.InclusionExclusion.InclusionExclusion import InclusionExclusion
from Dashboards.Telemetry.Telemetry import Telemetry
from Dashboards.Telemetry.ProductUsage import ProductUsage


class NikeCA(
    Snowflake
    , QA
    , Dashboards
    , InclusionExclusion
    , Telemetry
    , ProductUsage
):
    pass





class InclusionExclusion:

    def pos_data_quality_review(self
                                , username: str
                                , warehouse: str
                                , database: str
                                , role: str
                                , columns: list | None = None
                                , filters: list | None = None
                                , order_by: list | None = None
                                , polars: bool = False
                                , start_date: str | None = None
                                , end_date: str | None = None):

        from NikeCA import Snowflake
        import polars as pl
        import datetime
        import configparser

        config = configparser.ConfigParser()
        config.read('pip.ini')

        table = config['inclusion_exclusion_pos_data_quality_review'].get('table')
        a = config['inclusion_exclusion_summary_and_home'].get('col_a')

        # Ensure columns is a list
        if not (isinstance(columns, list) or columns is None):
            raise TypeError("columns must be a list")

        # Ensure filters is a list
        if not (isinstance(filters, list) or filters is None):
            raise TypeError("filters must be a list")

        # Ensure order_by is a list
        if not (isinstance(order_by, list) or order_by is None):
            raise TypeError("order_by must be a list")

        if columns is None:
            columns_str = config['inclusion_exclusion_pos_data_quality_review'].get('columns')
        else:
            columns_str = ', '.join(columns)

        if filters is None:
            filters = []
        # Join column filters into comma-separated string, if any
        if len(filters) > 0:
            filters_str = 'AND '.join(filters)
            filters_str = f'WHERE {filters_str}'
        else:
            filters_str = ''

        if order_by is None:
            order_by = config['inclusion_exclusion_pos_data_quality_review'].get('order_by')
        else:
            order_by = ', '.join(order_by)

        query = f"""
            SELECT
                {columns_str}
                
            FROM
                {table}
            {filters_str}
                
            ORDER BY
                {order_by}
                
            ;
        """

        df = pl.DataFrame(Snowflake(username=username, warehouse=warehouse, database=database,
                                    role=role).snowflake_pull(query=query, polars=polars))

        if start_date is None and end_date is None:
            return df.filter(pl.col(a) == pl.col(a).max())
        elif end_date is None:
            return df.filter((pl.col(a) >= datetime.datetime.strptime(start_date, '%Y-%m-%d')))
        elif start_date is None:
            return df.filter((pl.col(a) <= datetime.datetime.strptime(end_date, '%Y-%m-%d')))
        else:
            df = df.filter((pl.col(a) >= datetime.datetime.strptime(start_date, '%Y-%m-%d')))
            return df.filter((pl.col(a) <= datetime.datetime.strptime(end_date, '%Y-%m-%d')))

    def summary(self
                , username: str
                , warehouse: str
                , database: str
                , role: str
                , columns=None
                , filters=None
                , order_by=None
                , polars: bool = True
                , start_date: str | None = None
                , end_date: str | None = None
                ):

        import polars as pl
        import pandas as pd
        import configparser

        config = configparser.ConfigParser()
        config.read('pip.ini')

        a = config['inclusion_exclusion_summary_and_home'].get('col_a')
        b = config['inclusion_exclusion_summary_and_home'].get('col_b')
        c = config['inclusion_exclusion_summary_and_home'].get('col_c')

        df = InclusionExclusion.pos_data_quality_review(self, username=username, warehouse=warehouse,
                                                        database=database, role=role, columns=columns, filters=filters,
                                                        order_by=order_by, polars=polars, start_date=start_date,
                                                        end_date=end_date)

        df_pd = pd.DataFrame()

        df_pd['RETAILERS'] = df.select(pl.col(b).count()).to_arrow()[0]

        df_pd['INCLUDED'] = df.select(df.filter(pl.col(c) == 'Y').select(pl.col(b).count())).to_arrow()[0]
        df_pd['PERCENT_INCLUDED'] = f"{str(round(df_pd['INCLUDED']/df_pd['RETAILERS']*100, 0))}%"
        df_pd['EXCLUDED'] = df.select(df.filter(pl.col(c) == 'N').select(pl.col(b).count())).to_arrow()[0]
        df_pd['PERCENT_EXCLUDED'] = f"{str(round(df_pd['EXCLUDED']/df_pd['RETAILERS'] * 100, 0))}%"

        if polars:
            return pl.DataFrame(df_pd)
        return df_pd

    def home_summary(self
                     , username: str
                     , warehouse: str
                     , database: str
                     , role: str
                     , columns=None
                     , filters=None
                     , order_by=None
                     , polars: bool = True
                     , start_date: str | None = None
                     , end_date: str | None = None
                     ):

        import datetime

        import polars as pl
        import configparser

        config = configparser.ConfigParser()
        config.read('pip.ini')

        a = config['inclusion_exclusion_summary_and_home'].get('col_a')
        b = config['inclusion_exclusion_summary_and_home'].get('col_b')
        c = config['inclusion_exclusion_summary_and_home'].get('col_c')
        d = config['inclusion_exclusion_summary_and_home'].get('col_d')
        e = config['inclusion_exclusion_summary_and_home'].get('col_e')
        f = config['inclusion_exclusion_summary_and_home'].get('col_f')
        g = config['inclusion_exclusion_summary_and_home'].get('col_g')
        h = config['inclusion_exclusion_summary_and_home'].get('col_h')

        df = InclusionExclusion.pos_data_quality_review(self, username=username, warehouse=warehouse, database=database,
                                                        role=role, columns=columns, filters=filters, order_by=order_by,
                                                        polars=polars, start_date=start_date, end_date=end_date)
        df_count = (
            df
            .with_columns([
                pl.col(c).count().over([
                    d
                    ]).alias('Total Retailers')
                , pl.col(c).filter(
                    pl.col(c) == 'Y'
                    ).count().over(d).alias('INCLUDED')
                , pl.col(c).filter(
                    pl.col(c) == 'N'
                    ).count().over(d).alias('EXCLUDED')
                , pl.col(e).filter(
                    pl.col(e) == 'Y'
                    ).count().over(d).alias('SALE AVLB')
                , pl.col(e).filter(
                    pl.col(e) == 'N'
                    ).count().over(d).alias('SALE UNAVLB')
                , pl.col(f).filter(
                    pl.col(f) == 'Y'
                    ).count().over(d).alias('SALE IN TREND')
                , pl.col(f).filter(
                    pl.col(f) == 'N'
                    ).count().over(d).alias('SALE NOT IN TREND')
                , pl.col(g).filter(
                    pl.col(g) == 'Y'
                    ).count().over(d).alias('INV IN TREND')
                , pl.col(g).filter(
                    pl.col(g) == 'N'
                    ).count().over(d).alias('INV NOT IN TREND')


            ])
        ).select([h, 'Total Retailers', 'INCLUDED', 'EXCLUDED', 'SALE AVLB', 'SALE UNAVLB',
                  'SALE IN TREND', 'SALE NOT IN TREND', 'INV IN TREND', 'INV NOT IN TREND']).unique()

        df_count = (df_count
                    .lazy()
                    .with_columns([
                        (pl.col('INCLUDED') / pl.col('Total Retailers') * 100).round(0).alias('% INCLUDED')
                        , (pl.col('EXCLUDED') / pl.col('Total Retailers') * 100).round(0).alias('% EXCLUDED')
                        , (pl.col('SALE AVLB') / pl.col('Total Retailers') * 100).round(0).alias('% Retailers Sales Available')
                        , (pl.col('SALE UNAVLB') / pl.col('Total Retailers') * 100).round(0).alias('% Retailers Sales Unavailable')
                        , (pl.col('SALE IN TREND') / pl.col('Total Retailers') * 100).round(0).alias('% SALE IN TREND')
                        , (pl.col('SALE NOT IN TREND') / pl.col('Total Retailers') * 100).round(0).alias('% SALE NOT IN TREND')
                        , (pl.col('INV IN TREND') / pl.col('Total Retailers') * 100).round(0).alias('% INV IN TREND')
                        , (pl.col('INV NOT IN TREND') / pl.col('Total Retailers') * 100).round(0).alias('% INV NOT IN TREND')

        ]).collect()
                    )

        return df_count









class SnowflakePull:

    import pandas

    def snowflake_pull(self, query: str | dict | None, un, wh, db, role, schema=None, table=None,
                       sample_table: bool = False, sample_val: bool = False, table_sample: dict = None,
                       dtypes_conv=None, separate_dataframes: bool = True, polars: bool = False): # -> pandas.DataFrame:

        """
        function: pulls snowflake data

        dependencies: [
            pandas,
            snowflake.connector,
            time,
            datetime.datetime
        ]

        :param polars:
        :param separate_dataframes:
        :param table:
        :param schema:
        :param query: str | dict
            SQL query to run on Snowflake
                    E.G. query = "SELECT * FROM  NGP_DA_PROD.POS.TO_DATE_AGG_CHANNEL_CY"
            Can also be multiple queries in the form of a dictionary
                    E.G. query = {"df1": "SELECT * FROM  NGP_DA_PROD.POS.TO_DATE_AGG_CHANNEL_CY", "df "SELECT TOP 2 * \
                    FROM  NGP_DA_PROD.POS.TO_DATE_AGG_CHANNEL_CY"}

        :param un: str
            Nike Snowflake Username
                "USERNAME"

        :param db: str, default 'NA'
            Name of the Database

        :param wh: str
            Name of the Wharehouse
            e.g. "DA_DSM_SCANALYTICS_REPORTING_PROD"

        :param role: str
            Name of the role under which you are running Snowflake
                "DF_######"

        :param sample_table: bool, default: False

        :param sample_val: bool, default: False

        :param table_sample: dict, default: None
            later
                if table_sample = None
                    table_sample = {'db': None, 'schema': None, 'table': None, 'col': None}

        :param dtypes_conv: default: None

        :return: pandas.DataFrame
        """

        # snowflake connection packages:
        import pandas as pd
        import polars as pl
        import snowflake.connector
        import time

        if table_sample is not None:
            table_sample = {'db': None, 'schema': None, 'table': None, 'col': None}

        # --> take a random sample from a table in snowflake
        query = f'''SELECT * FROM {table_sample['db']}.{table_sample['schema']}.{table_sample['table']} LIMIT 100''' \
            if sample_table else query

        # --> take a random sample of a column from a table in snowflake
        query = f'''SELECT DISTINCT 
                {table_sample['col']} 
            FROM 
                {table_sample['db']}.{table_sample['schema']}.{table_sample['table']} 
            ORDER BY 1 LIMIT 10''' \
            if sample_val else query

        recs = False

        df = pd.DataFrame()

        if type(query) == dict:

            df = pd.DataFrame([query]).T
            df_index = df.index
            print(df_index)

            df2 = pl.from_pandas(df)
            print(df2)

            df_return = pd.DataFrame(index=df.index)
            df_return['sfqid'] = ''

            queries = len(df)
            print('Pulling ' + str(queries) + ' queries')

            query_list = []
            db_list = []
            complete = []

            for item in range(queries):
                query_list.append(item)
                db_list.append(item)
                complete.append(0)

            print('opening snowflake connection...')

            try:
                cnn = snowflake.connector.connect(
                    user=un,
                    account='nike',
                    authenticator='externalbrowser',
                    role=role,
                    warehouse='POS_REPORT_PROD'
                )

                cs = cnn.cursor()
                process_complete = 0
                process_pass = 0
                counter = 0

                for k, v in df.iterrows():
                    sql = v[0]
                    cs.execute_async(sql)
                    query_list[counter] = cs.sfqid
                    df_return['sfqid'][k] = cs.sfqid
                    counter += 1
                dfs = {}
                while process_complete == 0:
                    item = -1
                    process_pass += 1
                    if sum(complete) == queries or process_pass == 10:
                        process_complete = 1
                    for result in query_list:
                        item += 1
                        if complete[item] == 0:
                            status = cnn.get_query_status_throw_if_error(result)
                            print('the status for ' + df_return[df_return['sfqid'] == result].index[0] + ' is ' +
                                  str(status))
                            if str(status) == 'QueryStatus.SUCCESS':
                                complete[item] = 1
                                cs.get_results_from_sfqid(result)
                                if separate_dataframes:

                                    recs = True
                                    dfs[df_return[df_return['sfqid'] == result].index[0]] = cs.fetch_pandas_all()

                                else:
                                    df = pd.concat([df, cs.fetch_pandas_all()])
                            else:
                                time.sleep(.25)
            except Exception as e:
                print(e)
            finally:
                cnn.close()
                print('process complete')
        else:
            # connection settings
            from snowflake.connector.converter_null import SnowflakeNoConverterToPython
            conn = snowflake.connector.connect(
                user=un,
                account='nike',

                # opens separate browser window to confirm authentication
                authenticator='externalbrowser',
                warehouse=wh,
                database=db,
                role=role,
                converter_class=SnowflakeNoConverterToPython
            )

            # connect to snowflake using conn variables
            cur = conn.cursor()

            try:
                cur.execute(query)  # execute sql, store into-->

            # final data pull --> allows datatype-memory optimization
                df = pl.DataFrame
                if polars:
                    df = cur.fetch_arrow_all()

                else:
                    df = cur.fetch_pandas_all()
                print(df)

            except snowflake.connector.errors.ProgrammingError:
                print(f'Could not retrieve:\n\t {query}')

            except TypeError:
                print(f"""Could not retrieve:\n\t{query}\n\nbecause TypeError: NoneType""")

            # --> allows metadata querying
            except Exception as e:
                print(e)
                temp_df = cur.fetchall()  # return data
                cols = [x.name for x in cur.description]  # get column names
                df = pd.DataFrame(temp_df, columns=cols)  # create dataset

            conn.close()
            cur.close()
        if recs:
            return [dfs[k] for k in df_index]

        return df







class QA:

    def column_gap_analysis(self, df1,  df2, ds1_nm='Source #1', ds2_nm='Source #2', case_sens=True,
                            print_analysis=True, check_match_by=None, breakdown_grain=None):

        import time # timer package
        import pandas as pd
        import itertools

        # ERROR CHECKING #
        if (len(df1) == 0) | (len(df2) == 0):
            raise ValueError(f'''
            ERROR: one of the input data sources contains 0 rows of data
                Source #1: {len(df1)}
                Source #2: {len(df2)}
            ''')

        # other 'NULL' values to check for
        null_val_criteria = ['', ' ', '  ', '   ', 'N/A', 'NA', '*UNK*', 'UNKNOWN INTENDED USE', 'UNKNOWN DIMENSION',
                             '* DO NOT USE - TEST AFLSWNS SL', '*NUK*', 'UNKNOWN', None, '[]']

        # start timer
        tic = time.perf_counter()
        # --> uppercase all column names if case-sensitive flag is False
        if not case_sens:
            df1.columns = df1.columns.str.upper()
            df2.columns = df2.columns.str.upper()

        # data source 1 - # cols
        df1_cols = len(list(df1.columns))

        # data source 2 - # cols
        df2_cols = len(list(df2.columns))

        # matching cols
        match_cols = sorted([x for x in list(df1.columns) if x in (list(df2.columns))])

        # non-matching cols
        df1_non_matches = sorted([x for x in list(df1.columns) if x not in (list(df2.columns))])

        # non-matching cols
        df2_non_matches = sorted([x for x in list(df2.columns) if x not in (list(df1.columns))])

        # % of matching columns
        pct_match = 1 - ((len(df1_non_matches) + len(df2_non_matches)) / (df1_cols + df2_cols))

        # MATCHING COLUMNS ONLY --> determine coverage %
        results_df = pd.DataFrame(itertools.zip_longest(match_cols, df1_non_matches, df2_non_matches), columns=[
            'COL_MATCHES', f'{ds1_nm}_NON_MATCHES', f'{ds2_nm}_NON_MATCHES'])

        if len(match_cols) > 0:

            # --> perform analysis:

            # --> function: format dataframes for check_match_by flag = True analysis
            def format_df(inp_df, col_x):
                # drop any NA row
                f_df = inp_df[[col_x, check_match_by]].copy().dropna().reset_index(drop=True)

                # remove other NULL value rows
                f_df = f_df.loc[~(f_df[col_x].isin(null_val_criteria))].copy().reset_index(drop=True)

                # remove other NULL value rows
                f_df = f_df.loc[~(f_df[check_match_by].isin(null_val_criteria))].copy().reset_index(drop=True)

                # drop extra unneeded column
                f_df.drop(columns=[check_match_by], inplace=True)

                # drop duplicate records & reset index
                f_df = f_df.drop_duplicates().reset_index(drop=True)

                return f_df

            # --> function: compare
            def check_coverage(temp_df1, temp_df2, col_x, grain_brkdwn=None):

                # get ALL unique values in the matching columns
                if (col_x == check_match_by) | (check_match_by is None):

                    # total values in dataframe-column, excluding NA values
                    # ds1 unique values
                    ds1_uniq_val = pd.DataFrame(temp_df1.loc[~temp_df1[col_x].isna(),
                                                             col_x].copy()).drop_duplicates().reset_index(drop=True)

                    # ds2 unique values
                    ds2_uniq_val = pd.DataFrame(temp_df2.loc[~temp_df2[col_x].isna(),
                                                             col_x].copy()).drop_duplicates().reset_index(drop=True)

                # get ALL unique values in columns where the check_match_by column IS ALSO NOT NULL
                else:
                    ds1_uniq_val = format_df(temp_df1, col_x) # format dfs
                    ds2_uniq_val = format_df(temp_df2, col_x)

                try:
                    # ds1 values also in ds2
                    coverage_vals1 = ds1_uniq_val.merge(ds2_uniq_val, how='inner', on=x)

                    # ds2 values also in ds1
                    coverage_vals2 = ds2_uniq_val.merge(ds1_uniq_val, how='inner', on=x)
                except:
                    print(f'Dtype mismatch for {col_x}')
                    try: # --> attempt to convert datatypes & merge again

                        # ds1 values also in ds2
                        coverage_vals1 = ds1_uniq_val.merge(ds2_uniq_val.astype(ds1_uniq_val.dtypes[0]),
                                                            how='inner', on=x)

                        # ds2 values also in ds1
                        coverage_vals2 = ds2_uniq_val.merge(ds1_uniq_val.astype(ds2_uniq_val.dtypes[0]), how='inner',
                                                            on=x)
                    except:
                        raise ValueError \
                            (f'Could not fix Dtype mismatch on {col_x}: {ds1_uniq_val.dtypes[0]} vs '
                             f'{ds2_uniq_val.dtypes[0]} - fix formatting before running again')

                # calculate % of ds1 values covered by ds2
                # ds1 denominator
                ds1_den = len(ds1_uniq_val)

                # ds2 numerator
                ds2_num = len(coverage_vals1)

                # calculate % of ds2 values covered by ds1
                ds2_den = len(ds2_uniq_val) # ds1 denominator
                ds1_num = len(coverage_vals2) # ds2 numerator

                if grain_brkdwn is None:
                    try:

                        # append values to final list
                        val_holder.append([ds1_nm, ds2_nm, col_x, ds1_den, ds2_num, "{:.2%}".format(ds2_num / ds1_den)])
                        val_holder.append([ds2_nm, ds1_nm, col_x, ds2_den, ds1_num, "{:.2%}".format(ds1_num / ds2_den)])
                    except:
                        val_holder.append([ds1_nm, ds2_nm, col_x, ds1_den, ds2_num, "No pct"]) # append values to final list
                        val_holder.append([ds2_nm, ds1_nm, col_x, ds2_den, ds1_num, "No pct"])
                else:
                    try:
                        val_holder.append([ds1_nm, ds2_nm, col_x, grain_brkdwn, ds1_den, ds2_num, "{:.2%}".format(ds2_num / ds1_den)]) # append values to final list
                        val_holder.append \
                            ([ds2_nm, ds1_nm, col_x, grain_brkdwn, ds2_den, ds1_num, "{:.2%}".format(ds1_num / ds2_den)])
                    except:
                        val_holder.append([ds1_nm, ds2_nm, col_x, grain_brkdwn, ds1_den, ds2_num, "No pct"]) # append values to final list
                        val_holder.append([ds2_nm, ds1_nm, col_x, grain_brkdwn, ds2_den, ds1_num, "No pct"])


            val_holder = [] # data container
            for x in list(results_df.loc[~results_df['COL_MATCHES'].isna(), 'COL_MATCHES']):

                if breakdown_grain != None: # compare values by other grain
                    for val in (val for val in [*set(list(df1[breakdown_grain].unique()) + list(df2[breakdown_grain].unique()))] if val != None): # get all values from both lists
                        grain_df1 = df1.loc[df1[breakdown_grain] == val].copy().reset_index(drop=True)
                        grain_df2 = df2.loc[df2[breakdown_grain] == val].copy().reset_index(drop=True)
                        check_coverage(grain_df1, grain_df2, x, val)
                else:
                    check_coverage(df1, df2, x)

        # --> print analysis statement if specified
        if print_analysis == True:
            print(f'''
    ========================================================= DATA SOURCES ==========================================================
    Data Source #1: {ds1_nm}
    Data Source #2: {ds2_nm}
    ==================================================== DATAFRAME DESCRIPTIONS =====================================================
    (# rows, # cols)
    
    {ds1_nm} shape = {df1.shape}
    {ds1_nm} size = {df1.size}
    {ds2_nm} shape = {df2.shape}
    {ds2_nm} size = {df2.size}
    ======================================================== COLUMN MATCHES =========================================================
    
    Case sensitive? {case_sens}
    {len(match_cols)} matching column(s): {match_cols}
    {ds1_nm} non-matching column(s): {df1_non_matches}
    {ds2_nm} non-matching column(s): {df2_non_matches}
    ====================================================== PERCENTAGE MATCHING ======================================================
    Overall % column matches: {"{:.2%}".format(pct_match)} ({len(match_cols) * 2}/{(df1_cols + df2_cols)})
    {ds1_nm} % columns in {ds2_nm}:     {"{:.2%}".format(1 - (len(df1_non_matches) / df1_cols))} ({df1_cols - len(df1_non_matches)}/{df1_cols})
    {ds2_nm} % columns in {ds1_nm}:     {"{:.2%}".format(1 - (len(df2_non_matches) / df2_cols))} ({df2_cols - len(df2_non_matches)}/{df2_cols})
    Analysis time: {time.perf_counter() - tic:0.4f} seconds ({(time.perf_counter() - tic) / 60:0.4f} minutes)
            ''')

        try: # return column-value comparison if we have matching columns
            fin_cols = ['DS1_NAME', 'DS2_NAME', 'COLUMN', 'DS1_UNIQ_COL_VALS', 'DS2_MATCHING_VALS', 'DS2_COVERAGE_PCT'] if breakdown_grain == None else ['DS1_NAME', 'DS2_NAME', 'COLUMN', f'{breakdown_grain}_VALUE', 'DS1_UNIQ_COL_VALS', 'DS2_MATCHING_VALS', 'DS2_COVERAGE_PCT']
            return pd.DataFrame(data=val_holder, columns=fin_cols).sort_values(by=['DS1_NAME', 'COLUMN'], ascending=True).reset_index(drop=True)
        except: # return matching column results otherwise
            return results_df

    def data_prfl_analysis(self, df, ds_name='Data Source', sample_vals=5, print_analysis=True,
                           show_pct_fmt=True):  # --> sample values to take from each col

        import time  # timer package
        import pandas as pd
        import numpy as np
        from datetime import datetime

        if len(df) == 0:
            return pd.DataFrame(data=[],
                                columns=['DATA_SOURCE', 'COLUMN', 'COL_DATA_TYPE', 'TOTAL_ROWS', 'ROW_DTYPE_CT',
                                         'PRIMARY_DTYPE_PCT', 'COVERAGE_PCT', 'NULL_PCT',
                                         'DTYPE_ERROR_FLAG', 'NON_NULL_ROWS', 'NULL_VALUES', 'UNIQUE_VALUES',
                                         'COL_VALUE_SAMPLE', 'NULL_VALUE_SAMPLE'])

        print(f'Started data profiling query: {datetime.now().strftime("%H:%M:%S")}')
        tic = time.perf_counter()  # start timer
        analysis_keys = []  # --> stores information about columns in a data source
        null_val_criteria = ['', ' ', '  ', '   ', 'N/A', 'NA', '*UNK*', 'UNKNOWN INTENDED USE', 'UNKNOWN DIMENSION',
                             '* DO NOT USE - TEST AFLSWNS SL', '*NUK*', 'UNKNOWN', None,
                             '[]']  # other 'NULL' values to check for
        map_df = df.applymap(lambda x: type(x).__name__, na_action='ignore')  # dataframe data type mapping

        for col in list(df.columns):  # loop through each column in data source & get:
            total_col_vals = len(df[col].values)  # total rows in the column
            null_col_vals = len(
                df.loc[(df[col].isna()) | (df[col].isin(null_val_criteria))][col].values)  # NULL/unknown values only
            uniq_col_vals = len(np.unique(df.loc[~((df[col].isna()) | (df[col].isin(null_val_criteria)))][
                                              col].values))  # unique values, excluding NULL values
            row_dtypes = map_df[col].value_counts(
                sort=True).to_dict()  # --> check that row value data types match column data type
            max_dtype_val = max(row_dtypes.values()) if len(row_dtypes) > 0 else 0  # max values of 1 data type
            row_dtypes['NULL'] = null_col_vals  # add NULL values into dictionary
            dtype_error_flag = True if (null_col_vals + max_dtype_val) != total_col_vals else False
            null_list = list(df.loc[(df[col].isna()) | (df[col].isin(null_val_criteria))][col].unique())

            # add: column name, data type, # total rows, row datatypes, % primary dtype, % null values, # null rows,
            # unique values, unique value list, NULL value list
            analysis_keys.append([ds_name, col, df[col].dtype, total_col_vals, row_dtypes,
                                  "{:.2%}".format(max_dtype_val / total_col_vals) if show_pct_fmt else (
                                              max_dtype_val / total_col_vals),
                                  "{:.2%}".format(1 - (null_col_vals / total_col_vals)) if show_pct_fmt else (
                                              1 - (null_col_vals / total_col_vals)),
                                  "{:.2%}".format(null_col_vals / total_col_vals) if show_pct_fmt else (
                                              null_col_vals / total_col_vals),
                                  dtype_error_flag, total_col_vals - null_col_vals,
                                  null_col_vals, uniq_col_vals,
                                  sorted(list(np.unique(df.loc[~df[col].isna()][col].values))[:sample_vals]),
                                  'No NULLs' if len(null_list) == 0 else null_list])

        # return results dataframe analysis
        # --> COL_VALUE_SAMPLE includes 5 sample values (or more if specified) from the column
        results_df = pd.DataFrame(data=analysis_keys,
                                  columns=['DATA_SOURCE', 'COLUMN', 'COL_DATA_TYPE', 'TOTAL_ROWS', 'ROW_DTYPE_CT',
                                           'PRIMARY_DTYPE_PCT', 'COVERAGE_PCT', 'NULL_PCT',
                                           'DTYPE_ERROR_FLAG', 'NON_NULL_ROWS', 'NULL_VALUES', 'UNIQUE_VALUES',
                                           'COL_VALUE_SAMPLE', 'NULL_VALUE_SAMPLE'])
        dtype_counts = df.dtypes.value_counts()  # count of column datatypes
        dtype_statement = ''  # placeholder
        for i in list(dtype_counts.index):  # add a count for each datatype in the df
            dtype_statement += f'''({str(i)}: {str(dtype_counts[i])}); '''

        # --> function: get the # of columns with a % of NULL values >= input %
        def get_null_cols(pct):
            return len(results_df.loc[(results_df['NULL_VALUES'] / results_df['TOTAL_ROWS']) >= pct])

        analysis_statement = (f'''    
    * Finished data profiling for:      {ds_name}
    ============================================================================================================================================
    Total columns:                      {len(list(df.columns))}
    Sampled rows:                       {len(df)}
    ============================================================================================================================================
    Total 100% NULL value columns:      {get_null_cols(1)} ({"{:.2%}".format(get_null_cols(1) / len(list(df.columns)))} of columns)
        >=90% NULL value columns:       {get_null_cols(.9)} ({"{:.2%}".format(get_null_cols(.9) / len(list(df.columns)))} of columns)
        >=75% NULL value columns:       {get_null_cols(.75)} ({"{:.2%}".format(get_null_cols(.75) / len(list(df.columns)))} of columns)
        >=50% NULL value columns:       {get_null_cols(.5)} ({"{:.2%}".format(get_null_cols(.5) / len(list(df.columns)))} of columns)
        >=25% NULL value columns:       {get_null_cols(.25)} ({"{:.2%}".format(get_null_cols(.25) / len(list(df.columns)))} of columns)
        >=10% NULL value columns:       {get_null_cols(.1)} ({"{:.2%}".format(get_null_cols(.1) / len(list(df.columns)))} of columns)  
    Total >0% NULL value columns:       {get_null_cols(.00001)} ({"{:.2%}".format(get_null_cols(.00001) / len(list(df.columns)))} of columns)
    ============================================================================================================================================
    Column Data Types: {dtype_statement}
    ============================================================================================================================================ 
    # columns w/ non-matching dtype values:   {len(results_df.loc[results_df['DTYPE_ERROR_FLAG'] == True])} ({"{:.2%}".format(len(results_df.loc[results_df['DTYPE_ERROR_FLAG'] == True]) / len(list(df.columns)))} of columns)

        columns: {list(results_df.loc[results_df['DTYPE_ERROR_FLAG'] == True]['COLUMN'].unique())}     
    ''')  # --> final analysis

        print(f'''
    ============================================================================================================================================    
    Analysis time: {time.perf_counter() - tic:0.4f} seconds ({(time.perf_counter() - tic) / 60:0.4f} minutes)
    ============================================================================================================================================   
    ''')
        if print_analysis:
            print(analysis_statement)  # print analysis of results

        return results_df.sort_values(['COLUMN', 'UNIQUE_VALUES'], ascending=[True, False])

